#include <stdio.h>
#include <stdlib.h>
#include "TLBOCore.h"
#include "ConfigLoader.h"
#ifdef USE_MPI // The USE_MPI macro is set manually during compilation to enable the use of MPI
    #include <mpi.h>
#else
	#include "Timers.h"
#endif

#include "ObjectiveFunctions/Sphere.h" //Where to write specific objective functions

int main(int argc, char** argv){
    //Information of the objective function. Global updates here...
    double (*objFunc)(double*) = Sphere;
    int numVars = 0;
    double* bounds = 0;//These 2 variables should be written by the selected objective function
    writeInfo(&numVars, &bounds);//This kind of functions should be provided by the specific objective functions to be used
    double* solution = malloc(sizeof(double)*numVars);
    //TLBO Information that will be read. Launching procedure for MPI usage:
    int popSize = 0, cycles = 0, rnd_seed = 0;
    double val = 0.0; //Final value of the obtained solution
    #ifndef USE_MPI //No MPI-launching
        LoadConfiguration(argv[1], &popSize, &cycles, &rnd_seed, 1);//The last 1 tells the tool to display the loaded configuration
        srand(rnd_seed);//Setting the required seed for the master process (which defines TF, random values...)
        double t0 = get_wall_time();
        val = TLBOSearch(solution, numVars, objFunc, popSize, cycles, bounds, 0, popSize);//The last two parameters define the operative scope (useful for MPI processes)
        double t1 = get_wall_time();
        printf("Execution time (wall) (s): %lf\n", t1-t0);
        printf("\nFinal Value: %lf\n", val); printf("\n-.-.-.-.-.-.-.-.-.-.-.-.-\n");
        ReadMemory(solution, numVars);//Displaying the solution vector
    #else //MPI-launching
        int rank, teamSize;
        MPI_Init(&argc, &argv);//The name of the configuration file is expected by process 0
        MPI_Comm_size( MPI_COMM_WORLD, &teamSize );//Total number of MPI processes
        MPI_Comm_rank( MPI_COMM_WORLD, &rank );//Who am I?

        if(rank==0){//As the master thread I will load and share the global configuration
            LoadConfiguration(argv[1], &popSize, &cycles, &rnd_seed, 1);//The last 1 tells the tool to display the loaded configuration
        }
        MPI_Bcast(&popSize, 1, MPI_INT, 0, MPI_COMM_WORLD);//Sharing the population size
        MPI_Bcast(&cycles, 1, MPI_INT, 0, MPI_COMM_WORLD);//Sharing the number of cycles
        MPI_Bcast(&rnd_seed, 1, MPI_INT, 0, MPI_COMM_WORLD);//Sharing the random seed
        srand(rnd_seed + rank);//Every MPI-Process will have different seeds for future versions 
        //(Currently, all random values are generated by the master, other processes are mainly for share the total number of function evaluations per MPI node)
        //Let's start launching:
        int chunkSize = popSize/teamSize;
        int chunkOffset = popSize % teamSize;
        if(rank==0){
			double t0 = MPI_Wtime();
            val = TLBOSearch(solution, numVars, objFunc, popSize, cycles, bounds, 0, chunkSize + (chunkOffset>0));
            double t1 = MPI_Wtime();
            printf("Execution time (wall) (s): %lf\n", t1-t0);
            printf("\nFinal Value: %lf\n", val);
            //ReadMemory(solution, numVars);//Displaying the solution vector
        }else{
            int start = rank*chunkSize + ((rank>=chunkOffset)?chunkOffset:rank);//ranges are expected to be [,)
            int end = start + chunkSize + (rank<chunkOffset);
            val = TLBOSearch(solution, numVars, objFunc, popSize, cycles, bounds, start, end);
        }
        MPI_Finalize();
    #endif
    free(solution);//Clean-up
    free(bounds);
    return 0;
}